import argparse
import ast
from collections import namedtuple

from .formatters import format_choices


class IdValuePair(argparse.Action):

    def __init__(
            self,
            id_choices,
            value_type,
            value_metavar,
            sub_action='store',
            **kwargs,
        ):
        super().__init__(nargs=2, **kwargs)
        self.id_choices = id_choices
        self.value_type = value_type
        if sub_action not in ('store', 'append'):
            raise ValueError(f'unknown sub_action {sub_action:r}')
        self.sub_action = sub_action
        self.metavar = (format_choices(id_choices), value_metavar)

    def __call__(self, parser, namespace, values, option_string=None):
        id_, value = values
        try:
            value = self.value_type(value)
        except (TypeError, ValueError):
            raise argparse.ArgumentError(
                self,
                f"invalid {self.value_type.__name__}: {value!r}",
            )

        if id_ not in self.id_choices:
            raise argparse.ArgumentError(
                self,
                f"invalid choice: '{id_}' "
                f"(choose from {', '.join(map(repr, self.id_choices))})",
            )

        if self.sub_action == 'store':
            setattr(namespace, self.dest, (id_, value))
        elif self.sub_action == 'append':
            if not getattr(namespace, self.dest, None):
                setattr(namespace, self.dest, [])
            getattr(namespace, self.dest).append((id_, value))
        else:
            raise AssertionError


class IdKwargs(argparse.Action):

    IdKwargsPair = namedtuple('IdKwargsPair', ['id', 'kwargs'])

    def __init__(
            self,
            id_choices,
            split_token=',',
            use_bool_abbreviation=True,
            default_as_string=True,
            sub_action='store',
            **kwargs,
        ):
        super().__init__(nargs='+', **kwargs)
        self.id_choices = id_choices
        self.split_token = split_token
        self.default_as_string = default_as_string
        self.use_bool_abbreviation = use_bool_abbreviation
        if sub_action not in ('store', 'append'):
            raise ValueError(f'unknown sub_action {sub_action:r}')
        self.sub_action = sub_action
        self.metavar = (format_choices(id_choices), f'KEY1=VALUE1{split_token}KEY2=VALUE2')

    def __call__(self, parser, namespace, values, option_string=None):
        if len(values) == 2:
            id_, kwarg_string = values
            try:
                kwargs = self._process_kwargs_string(kwarg_string)
            except ValueError:
                raise argparse.ArgumentError(
                    self,
                    f"invalid kwargs: {kwarg_string!r}",
                )
            except NameError:
                raise argparse.ArgumentError(
                    self,
                    "value should be built-in types.",
                )
        elif len(values) == 1:
            id_, kwargs = values[0], {}
        else:
            raise argparse.ArgumentError(
                self,
                'expected at most 2 argument',
            )

        if id_ not in self.id_choices:
            raise argparse.ArgumentError(
                self,
                f"invalid choice: '{id_}' "
                f"(choose from {', '.join(map(repr, self.id_choices))})",
            )

        if self.sub_action == 'store':
            setattr(namespace, self.dest, self.IdKwargsPair(id_, kwargs))
        elif self.sub_action == 'append':
            if not getattr(namespace, self.dest, None):
                setattr(namespace, self.dest, [])
            getattr(namespace, self.dest).append(self.IdKwargsPair(id_, kwargs))
        else:
            raise AssertionError

    def _process_kwargs_string(self, kwarg_string):
        kwargs = {}
        kvs = kwarg_string.split(self.split_token)
        for kv in kvs:
            if '=' in kv:
                key, val = kv.split('=')
                try:
                    val = ast.literal_eval(val)
                except (SyntaxError, ValueError):
                    if not self.default_as_string:
                        raise
                    # default as string if can't eval
            elif self.use_bool_abbreviation:
                key, val = kv, True
            else:
                raise ValueError

            if key in kwargs:
                raise ValueError
            kwargs[key] = val

        return kwargs



import os
import pytest

from ..types import (
    int_in_range,
    float_in_range,
    path,
    filepath,
    dirpath,
)


@pytest.mark.parametrize('func, x, valid', [
    (int_in_range(0), '2', True),
    (int_in_range(0), '0', True),
    (int_in_range(2, 5), '6', False),
    (int_in_range(), 'a', False),
    (int_in_range(), '2.', False),
])
def test_int_in_range(func, x, valid):
    if valid:
        assert func(x) == int(x)
    else:
        with pytest.raises(ValueError):
            func(x)


@pytest.mark.parametrize('func, x, valid', [
    (float_in_range(0), '2', True),
    (float_in_range(0), '0', True),
    (float_in_range(0, inclusive=False), '0', False),
    (float_in_range(), 'a', False),
    (float_in_range(), 'inf', False),
])
def test_float_in_range(func, x, valid):
    if valid:
        assert func(x) == float(x)
    else:
        with pytest.raises(ValueError):
            func(x)


def test_path():
    assert path("A//B") == path("A/B/") == path("A/./B") == path("A/foo/../B") == "A/B"


def test_filepath(tmpdir):
    filename = os.path.join(tmpdir, 'new_file')
    with pytest.raises(ValueError):
        filepath(filename)

    with open(filename, 'w'):
        pass
    assert filepath(filename) == filename


def test_dirpath(tmpdir):
    assert dirpath(tmpdir) == tmpdir
    with pytest.raises(ValueError):
        dirpath(os.path.join(tmpdir, 'dir_not_existed'))


@pytest.mark.parametrize('func, expected_name', [
    (int_in_range(0), 'nonnegative_int'),
    (int_in_range(1), 'positive_int'),
    (int_in_range(2), 'int∈[2, ∞)'),
    (int_in_range(2, 5), 'int∈[2, 5]'),
    (float_in_range(0.), 'nonnegative_float'),
    (float_in_range(0., inclusive=False), 'positive_float'),
    (float_in_range(1.), 'float∈[1.0, ∞)'),
    (float_in_range(1., inclusive=False), 'float∈(1.0, ∞)'),
])
def test_name(func, expected_name):
    assert func.__name__ == expected_name



import math
import os


def int_in_range(minval=-math.inf, maxval=math.inf):

    def func(x):
        return _validate_in_range(int(x), minval, maxval, inclusive=True)

    if (minval, maxval) == (1, math.inf):
        func.__name__ = 'positive_int'
    elif (minval, maxval) == (0, math.inf):
        func.__name__ = 'nonnegative_int'
    else:
        func.__name__ = f'int∈{_repr_inteval(minval, maxval, inclusive=True)}'
    return func


def float_in_range(minval=-math.inf, maxval=math.inf, inclusive=True):

    def func(x):
        return _validate_in_range(float(x), minval, maxval, inclusive=inclusive)

    if (minval, maxval, inclusive) == (0., math.inf, False):
        func.__name__ = 'positive_float'
    elif (minval, maxval, inclusive) == (0., math.inf, True):
        func.__name__ = 'nonnegative_float'
    else:
        func.__name__ = f'float∈{_repr_inteval(minval, maxval, inclusive)}'
    return func


def _validate_in_range(x, minval, maxval, inclusive):
    if math.isinf(x):
        raise ValueError
    if inclusive:
        if not (minval <= x <= maxval):
            raise ValueError
    elif not (minval < x < maxval):
        raise ValueError
    return x


def _repr_inteval(minval, maxval, inclusive):

    def _math_repr(x):
        if x == -math.inf:
            return '-∞'
        if x == math.inf:
            return '∞'
        return repr(x)

    left_bracket = '[' if (inclusive and not math.isinf(minval)) else '('
    right_bracket = ']' if (inclusive and not math.isinf(maxval)) else ')'
    return f"{left_bracket}{_math_repr(minval)}, {_math_repr(maxval)}{right_bracket}"


def path(x):
    return os.path.normpath(x)


def filepath(x):
    """Check if file exists.
    """
    x = path(x)
    if not os.path.isfile(x):
        raise ValueError
    return x


def dirpath(x):
    """Check if directory exists.
    """
    x = path(x)
    if not os.path.isdir(x):
        raise ValueError
    return x



import argparse
import re
import shutil
import termcolor


class YoctolFormatter(argparse.ArgumentDefaultsHelpFormatter):

    def __init__(self, prog):
        super().__init__(prog, max_help_position=4, width=shutil.get_terminal_size()[0])

    # HACK Override: Change some format part to avoid [], () bracket broken
    # https://github.com/python/cpython/blob/3.7/Lib/argparse.py#L391-L486
    def _format_actions_usage(self, actions, groups):  # noqa: C901
        ########## COPY FROM SOURCE CODE ##########
        # find group indices and identify actions in groups
        group_actions = set()
        inserts = {}
        for group in groups:
            try:
                start = actions.index(group._group_actions[0])
            except ValueError:
                continue
            else:
                end = start + len(group._group_actions)
                if actions[start:end] == group._group_actions:
                    for action in group._group_actions:
                        group_actions.add(action)
                    if not group.required:
                        if start in inserts:
                            inserts[start] += ' ['
                        else:
                            inserts[start] = '['
                        inserts[end] = ']'
                    else:
                        if start in inserts:
                            inserts[start] += ' ('
                        else:
                            inserts[start] = '('
                        inserts[end] = ')'
                    for i in range(start + 1, end):
                        inserts[i] = '|'

        # collect all actions format strings
        parts = []
        for i, action in enumerate(actions):

            # suppressed arguments are marked with None
            # remove | separators for suppressed arguments
            if action.help is argparse.SUPPRESS:
                parts.append(None)
                if inserts.get(i) == '|':
                    inserts.pop(i)
                elif inserts.get(i + 1) == '|':
                    inserts.pop(i + 1)

            # produce all arg strings
            elif not action.option_strings:
                default = self._get_default_metavar_for_positional(action)
                part = self._format_args(action, default)

                # if it's in a group, strip the outer []
                if action in group_actions:
                    if part[0] == '[' and part[-1] == ']':
                        part = part[1:-1]

                # add the action string to the list
                parts.append(part)

            # produce the first way to invoke the option in brackets
            else:
                option_string = action.option_strings[0]

                # if the Optional doesn't take a value, format is:
                #    -s or --long
                if action.nargs == 0:
                    part = '%s' % option_string

                # if the Optional takes a value, format is:
                #    -s ARGS or --long ARGS
                else:
                    default = self._get_default_metavar_for_optional(action)
                    args_string = self._format_args(action, default)
                    part = '%s %s' % (option_string, args_string)

                # make it look optional if it's not required or in a group
                if not action.required and action not in group_actions:
                    part = '[%s]' % part

                # add the action string to the list
                parts.append(part)

        # insert things at the necessary indices
        for i in sorted(inserts, reverse=True):
            parts[i:i] = [inserts[i]]

        # join all the action items with spaces
        text = ' '.join([item for item in parts if item is not None])

        # clean up separators for mutually exclusive groups
        open_ = r'[\[(]'
        close = r'[\])]'
        text = re.sub(r'(%s) ' % open_, r'\1', text)
        text = re.sub(r' (%s)' % close, r'\1', text)
        text = re.sub(r'%s *%s' % (open_, close), r'', text)
        ########## END COPY FROM SOURCE CODE ##########
        # remove L482 from source code to avoid [], () bracket broken
        text = text.strip()

        # return the text
        return text

    def _metavar_formatter(self, action, default_metavar):
        if action.metavar is not None:
            result = action.metavar
        elif action.choices is not None:
            # custom format for choices
            result = format_choices(action.choices)
        else:
            result = default_metavar

        def format_func(tuple_size):
            if isinstance(result, tuple):
                return result
            else:
                return (result, ) * tuple_size
        return format_func

    def _get_default_metavar_for_optional(self, action):
        # add default if action.type has no '__name__' attribute
        return getattr(action.type, '__name__', repr(action.type))

    def _get_default_metavar_for_positional(self, action):
        # add default if action.type has no '__name__' attribute
        return getattr(action.type, '__name__', repr(action.type))


def format_choices(choices):
    choice_strs = [
        termcolor.colored(str(choice), attrs=['underline', 'dark'])
        for choice in choices
    ]
    return f"{{{', '.join(choice_strs)}}}"



import argparse

from .formatters import YoctolFormatter


class YoctolArgumentParser(argparse.ArgumentParser):

    def __init__(
            self,
            prog=None,
            usage=None,
            description=None,
            epilog=None,
            parents=(),
            formatter_class=YoctolFormatter,
            prefix_chars='-',
            fromfile_prefix_chars=None,
            argument_default=None,
            conflict_handler='error',
            add_help=True,
            allow_abbrev=True,
        ):
        super().__init__(
            prog=prog,
            usage=usage,
            description=description,
            epilog=epilog,
            parents=parents,
            formatter_class=formatter_class,
            prefix_chars=prefix_chars,
            fromfile_prefix_chars=fromfile_prefix_chars,
            argument_default=argument_default,
            conflict_handler=conflict_handler,
            add_help=add_help,
            allow_abbrev=allow_abbrev,
        )



from .__version__ import (
    __title__,
    __version__,
    __description__,
    __author__,
)

from .actions import IdValuePair, IdKwargs
from .formatters import YoctolFormatter
from .parser import YoctolArgumentParser
from .types import (
    int_in_range,
    float_in_range,
    path,
    filepath,
    dirpath,
)

# just for import sugar
from argparse import (
    FileType,
    Namespace,
    Action,
    ONE_OR_MORE,
    OPTIONAL,
    REMAINDER,
    SUPPRESS,
    ZERO_OR_MORE,
)



import os
from pathlib import Path

from setuptools import setup, find_packages


readme = Path(__file__).parent.joinpath('README.md')
if readme.exists():
    with readme.open() as f:
        long_description = f.read()
else:
    long_description = '-'


here = os.path.abspath(os.path.dirname(__file__))
about = {}
with open(os.path.join(here, 'yoctol_argparse', '__version__.py'), 'r') as filep:
    exec(filep.read(), about)


setup(
    name=about['__title__'],
    version=about['__version__'],
    description=about['__description__'],
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires='>=3.6',
    packages=find_packages(),
    author=about['__author__'],
    author_email='jsaon@yoctol.com',
    url='',
    license='MIT',
    install_requires=[],
)



import argparse
import pytest
from unittest.mock import patch

from ..parser import YoctolArgumentParser


@pytest.mark.parametrize('arg', [
    'main.py -h',
    'main.py --help',
])
def test_help_formatter(arg):
    parser = YoctolArgumentParser()

    parser.add_argument('x')
    parser.add_argument('--y', help=argparse.SUPPRESS)
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--foo', choices=['a', 'b', 'c'], action='append')
    group.add_argument('--boo', choices=['a', 'b', 'c'], action='append')

    with patch('sys.argv', arg.split()), pytest.raises(SystemExit) as exc_info:
        parser.parse_args()
    assert exc_info.value.code == 0



import pytest
from unittest.mock import patch

from ..actions import IdValuePair, IdKwargs
from ..parser import YoctolArgumentParser


class TestIdValuePair:

    @pytest.fixture(scope='class')
    def yoctol_parser(self):
        parser = YoctolArgumentParser(prog='main.py')
        parser.add_argument(
            '--foo',
            action=IdValuePair,
            id_choices=['a', 'b'],
            value_metavar='boo',
            value_type=int,
            sub_action='append',
        )
        return parser

    def test_append(self, yoctol_parser):
        with patch('sys.argv', 'main.py --foo a 1 --foo b 2 --foo a 3'.split(' ')):
            args = yoctol_parser.parse_args()
        assert args.foo == [('a', 1), ('b', 2), ('a', 3)]

    @pytest.mark.parametrize('invalid_arg', [
        pytest.param('main.py --foo a', id='invalid_nargs'),
        pytest.param('main.py --foo a 1 b', id='invalid_nargs'),
        pytest.param('main.py --foo c 1', id='invalid_choice'),
        pytest.param('main.py --foo a b', id='invalid_value'),
    ])
    def test_raise_invalid_arg(self, yoctol_parser, invalid_arg):
        argv = invalid_arg.split()
        with patch('sys.argv', argv), pytest.raises(SystemExit) as exc_info:
            yoctol_parser.parse_args()
        assert exc_info.value.code == 2


class TestStoreIdKwargs:

    @pytest.fixture(scope='class')
    def yoctol_parser(self):
        parser = YoctolArgumentParser(prog='main.py')
        parser.add_argument(
            '--foo',
            action=IdKwargs,
            id_choices=['a', 'b'],
            use_bool_abbreviation=True,
            sub_action='store',
        )
        return parser

    @pytest.mark.parametrize('arg_string, expected_output', [
        ['main.py --foo a', ('a', {})],
        [
            'main.py --foo a I=1,F1=1.,F2=1e-4,F3=-1e-4',
            ('a', {'I': 1, 'F1': 1., 'F2': 1e-4, 'F3': -1e-4}),
        ],
        [
            'main.py --foo a B1=False,B2=True,B3,N=None',
            ('a', {'B1': False, 'B2': True, 'B3': True, 'N': None}),
        ],
        [
            'main.py --foo a S1=s,S2="s",S3=\'s\'',
            ('a', {'S1': 's', 'S2': 's', 'S3': 's'}),
        ],
        [
            'main.py --foo a 1=open,2=exit,3=exec,4=import,5=OSError',  # no builtins allowed
            ('a', {'1': 'open', '2': 'exit', '3': 'exec', '4': 'import', '5': 'OSError'}),
        ],
    ])
    def test_store(self, yoctol_parser, arg_string, expected_output):
        with patch('sys.argv', arg_string.split(' ')):
            args = yoctol_parser.parse_args()
        assert args.foo == expected_output

    @pytest.mark.parametrize('invalid_arg', [
        pytest.param('main.py --foo', id='nargs<1'),
        pytest.param('main.py --foo a 1 b', id='nargs>2'),
        pytest.param('main.py --foo c 1', id='invalid_choice'),
        pytest.param('main.py --foo a x=1=y', id='invalid_format_='),
        pytest.param('main.py --foo a x=1+y=y', id='invalid_format_split'),
        pytest.param('main.py --foo a x=1,x=2', id='duplicated_key'),
    ])
    def test_raise_invalid_arg(self, yoctol_parser, invalid_arg):
        argv = invalid_arg.split()
        with patch('sys.argv', argv), pytest.raises(SystemExit) as exc_info:
            yoctol_parser.parse_args()
        assert exc_info.value.code == 2
